! ::: -----------------------------------------------------------
! ::: This routine will tag high error cells based on the state array
! :::
! ::: INPUTS/OUTPUTS:
! :::
! ::: tag      <=  integer tag array
! ::: lo,hi     => index extent of tag array
! ::: set       => integer value to tag cell for refinement
! ::: clear     => integer value to untag cell
! ::: var       => array of data
! ::: nd        => number of components in var array (should be 1)
! ::: domlo,hi  => index extent of problem domain
! ::: delta     => cell spacing
! ::: xlo       => physical location of lower left hand
! :::              corner of tag array
! ::: problo    => phys loc of lower left corner of prob domain
! ::: time      => problem evolution time
! ::: level     => refinement level of this array
! ::: -----------------------------------------------------------
      subroutine ca_state_error(tag,tagl1,tagl2,tagl3,tagh1,tagh2,tagh3, &
                                set,clear, &
                                var,varl1,varl2,varl3,varh1,varh2,varh3, &
                                lo,hi,nd,domlo,domhi, &
                                delta,xlo,problo,time,level)
      use probdata_module
      implicit none

      integer          :: set, clear, nd, level
      integer          :: tagl1,tagl2,tagl3,tagh1,tagh2,tagh3
      integer          :: varl1,varl2,varl3,varh1,varh2,varh3
      integer          :: lo(3), hi(3), domlo(3), domhi(3)
      integer          :: tag(tagl1:tagh1,tagl2:tagh2,tagl3:tagh3)
      double precision :: var(varl1:varh1,varl2:varh2,varl3:varh3,nd)
      double precision :: delta(3), xlo(3), problo(3), time

      double precision :: den,temp,c12,ax,ay,az
      integer          :: i,j,k

      ! DENSITY     is component 1
      ! TEMPERATURE is component 2
      ! CARBON FRAC is component 3
 
!     Tag on regions of high density
      if (level .lt. max_denerr_lev) then
         if (level .eq. 0) then
            !$OMP PARALLEL DO PRIVATE(i,j,k,den)
            do k = lo(3), hi(3)
            do j = lo(2), hi(2)
            do i = lo(1), hi(1)
               den = var(i,j,k,1)
               if (den .ge. denerr0)then
                  tag(i,j,k) = set
               endif
            enddo
            enddo
            enddo
            !$OMP END PARALLEL DO
         else if (level .eq. 1) then
            !$OMP PARALLEL DO PRIVATE(i,j,k,den)
            do k = lo(3), hi(3)
            do j = lo(2), hi(2)
            do i = lo(1), hi(1)
               den = var(i,j,k,1)
               if (den .ge. denerr1)then
                  tag(i,j,k) = set
               endif
            enddo
            enddo
            enddo
            !$OMP END PARALLEL DO
         else if (level .eq. 2) then
            !$OMP PARALLEL DO PRIVATE(i,j,k,den)
            do k = lo(3), hi(3)
            do j = lo(2), hi(2)
            do i = lo(1), hi(1)
               den = var(i,j,k,1)
               if (den .ge. denerr2)then
                  tag(i,j,k) = set
               endif
            enddo
            enddo
            enddo
            !$OMP END PARALLEL DO
         else if (level .ge. 3) then
            !$OMP PARALLEL DO PRIVATE(i,j,k,den)
            do k = lo(3), hi(3)
            do j = lo(2), hi(2)
            do i = lo(1), hi(1)
               den = var(i,j,k,1)
               if (den .ge. denerr3)then
                  tag(i,j,k) = set
               endif
            enddo
            enddo
            enddo
            !$OMP END PARALLEL DO
         end if
      endif

!     Tag on regions of high temperature
      if (level .lt. max_temperr_lev) then
         !$OMP PARALLEL DO PRIVATE(i,j,k,den,temp)
         do k = lo(3), hi(3)
         do j = lo(2), hi(2)
         do i = lo(1), hi(1)
            den  = var(i,j,k,1)
            temp = var(i,j,k,2)
            if (temp .ge. temperr .and. den.ge.denthreshold) then
               tag(i,j,k) = set
            endif
         enddo
         enddo
         enddo
         !$OMP END PARALLEL DO
      endif

!     Tag on regions of low carbon fraction
      if (level .lt. max_specerr_lev) then
         !$OMP PARALLEL DO PRIVATE(i,j,k,den,temp,c12)
         do k = lo(3), hi(3)
         do j = lo(2), hi(2)
         do i = lo(1), hi(1)
            den  = var(i,j,k,1)
            temp = var(i,j,k,2)
            c12  = var(i,j,k,3)
            if (c12 .lt. specerr .and. den.ge.denthreshold) then
!            if (c12 .ge. 0.01d0 .and. c12 .lt. 0.49d0) then
               tag(i,j,k) = set
            endif
         enddo
         enddo
         enddo
         !$OMP END PARALLEL DO
      endif

!     Tag on regions of carbon fraction gradient 
      if (level .lt. max_specgrad_lev) then
         !$OMP PARALLEL DO PRIVATE(i,j,k,den,ax,ay,az)
         do k = lo(3), hi(3)
         do j = lo(2), hi(2)
         do i = lo(1), hi(1)
            den  = var(i,j,k,1)
            ax = ABS(var(i+1,j,k,3) - var(i,j,k,3))
            ay = ABS(var(i,j+1,k,3) - var(i,j,k,3))
            az = ABS(var(i,j,k+1,3) - var(i,j,k,3))
            ax = MAX(ax,ABS(var(i,j,k,3) - var(i-1,j,k,3)))
            ay = MAX(ay,ABS(var(i,j,k,3) - var(i,j-1,k,3)))
            az = MAX(az,ABS(var(i,j,k,3) - var(i,j,k-1,3)))
            if ( MAX(MAX(ax,ay),az) .ge. specgrad .and. den.ge.0.1d0*denthreshold) then
               tag(i,j,k) = set
            endif
         enddo
         enddo
         enddo
         !$OMP END PARALLEL DO
      endif

      end subroutine ca_state_error

